///////////////////////////////////////////////////////////////////////////////
// Project:     M
// File name:   gui/wxMenuDefs.cpp - definitions of al application menus
// Purpose:     gathering all the menus in one place makes it easier to find
//              and modify them
// Author:      Vadim Zeitlin
// Modified by:
// Created:     06.08.98
// CVS-ID:      $Id: wxMenuDefs.cpp,v 1.49 1999/06/23 12:52:54 VZ Exp $
// Copyright:   (c) 1998 Vadim Zeitlin <zeitlin@dptmaths.ens-cachan.fr>
// Licence:     M license
///////////////////////////////////////////////////////////////////////////////

// ============================================================================
// declarations
// ============================================================================

// ----------------------------------------------------------------------------
// headers
// ----------------------------------------------------------------------------
#include "Mpch.h"
#include "Mcommon.h"

#ifndef USE_PCH
# include "PathFinder.h"

# include "MApplication.h"
# include "gui/wxMApp.h"

# include "guidef.h"
#endif

#include <wx/menu.h>
#include <wx/intl.h>

#include "gui/wxIconManager.h"

#include "gui/wxMenuDefs.h"

// ----------------------------------------------------------------------------
// macros
// ----------------------------------------------------------------------------

#ifdef OS_WIN
# define BMP(name) (mApplication->GetIconManager()->GetBitmap(name))
#else
# define BMP(name)  ICON(name)
#endif

/// convenience macro to add a button to a toolbar
#define   TB_AddTool(tb, bmp, id, helptext) \
    tb->AddTool(id, BMP(bmp), wxNullBitmap, FALSE, -1, -1, NULL, _(helptext))

// ----------------------------------------------------------------------------
// local data
// ----------------------------------------------------------------------------

// structure describing a menu item
struct MenuItemInfo
{
   int idMenu;
   const char *label;
   const char *helpstring;
};

struct TbarItemInfo
{
   const char *bmp;      // image file or ressource name
   int         id;       // id of the associated command (-1 => separator)
   const char *tooltip;  // flyby text
};

// array of descriptions of all toolbar buttons, should be in sync with the enum
// in wxMenuDefs.h!
static const TbarItemInfo g_aToolBarData[] =
{
   // separator
   { "", -1, "" },

   // common for all frames
   { "tb_close",       WXMENU_FILE_CLOSE,  gettext_noop("Close this window") },
   { "tb_adrbook",    WXMENU_EDIT_ADB,    gettext_noop("Edit address book") },
   { "tb_preferences", WXMENU_EDIT_PREF,   gettext_noop("Edit preferences")  },

   // main frame
   { "tb_open",          WXMENU_FILE_OPEN,     gettext_noop("Open folder")           },
   { "tb_mail_compose",  WXMENU_FILE_COMPOSE,  gettext_noop("Compose a new message") },
   { "tb_help",          WXMENU_HELP_CONTEXT,  gettext_noop("Help")                  },
   { "tb_exit",          WXMENU_FILE_EXIT,     gettext_noop("Exit the program")      },

   // compose frame
   { "tb_print",     WXMENU_COMPOSE_PRINT,     gettext_noop("Print")         },
   { "tb_new",       WXMENU_COMPOSE_CLEAR,     gettext_noop("Clear Window")  },
   { "tb_attach",    WXMENU_COMPOSE_INSERTFILE,gettext_noop("Insert File")   },
   { "tb_editor",    WXMENU_COMPOSE_EXTEDIT,   gettext_noop("Call external editor")   },
   { "tb_mail_send", WXMENU_COMPOSE_SEND,      gettext_noop("Send Message")  },

   // folder and message view frames
   { "tb_mail",          WXMENU_MSG_OPEN,      gettext_noop("Open message")      },
   { "tb_mail_forward",  WXMENU_MSG_FORWARD,   gettext_noop("Forward message")   },
   { "tb_mail_reply",    WXMENU_MSG_REPLY,     gettext_noop("Reply to message")  },
   { "tb_print",         WXMENU_MSG_PRINT,     gettext_noop("Print message")     },
   { "tb_trash",         WXMENU_MSG_DELETE,    gettext_noop("Delete message")    },

   // ADB edit frame
   //{ "tb_open",     WXMENU_ADBBOOK_OPEN,    gettext_noop("Open address book file")  },
   { "tb_new",      WXMENU_ADBEDIT_NEW,     gettext_noop("Create new entry")        },
   { "tb_delete",   WXMENU_ADBEDIT_DELETE,  gettext_noop("Delete")                  },
   { "tb_undo",     WXMENU_ADBEDIT_UNDO,    gettext_noop("Undo")                    },
   { "tb_lookup",   WXMENU_ADBFIND_NEXT,    gettext_noop("Find next")               },
};

// arrays containing tbar buttons for each frame (must be -1 terminated!)
// the "Close", "Help" and "Exit" icons are added to all frames (except that
// "Close" is not added to the main frame because there it's the same as "Exit")
// @@ should we also add "Edit adb"/"Preferences" to all frames by default?

static const int g_aMainTbar[] =
{
   WXTBAR_MAIN_OPEN,
   WXTBAR_MAIN_COMPOSE,
   WXTBAR_SEP,
   WXTBAR_ADB,
   WXTBAR_PREFERENCES,
   -1
};

static const int g_aComposeTbar[] =
{
   WXTBAR_COMPOSE_PRINT,
   WXTBAR_COMPOSE_CLEAR,
   WXTBAR_COMPOSE_INSERT,
   WXTBAR_COMPOSE_EXTEDIT,
   WXTBAR_SEP,
   WXTBAR_COMPOSE_SEND,
   WXTBAR_SEP,
   WXTBAR_ADB,
   WXTBAR_PREFERENCES,
   -1
};

static const int g_aFolderTbar[] =
{
   WXTBAR_MAIN_OPEN,
   WXTBAR_MSG_OPEN,
   WXTBAR_SEP,
   WXTBAR_MAIN_COMPOSE,
   WXTBAR_MSG_FORWARD,
   WXTBAR_MSG_REPLY,
   WXTBAR_MSG_PRINT,
   WXTBAR_MSG_DELETE,
   WXTBAR_SEP,
   WXTBAR_ADB,
   WXTBAR_PREFERENCES,
   -1
};

static const int g_aMsgTbar[] =
{
   WXTBAR_MAIN_OPEN,
   WXTBAR_SEP,
   WXTBAR_MAIN_COMPOSE,
   WXTBAR_MSG_FORWARD,
   WXTBAR_MSG_REPLY,
   WXTBAR_MSG_PRINT,
   WXTBAR_MSG_DELETE,
   WXTBAR_SEP,
   WXTBAR_ADB,
   WXTBAR_PREFERENCES,
   -1
};

static const int g_aAdbTbar[] =
{
   WXTBAR_SEP,
   WXTBAR_ADB_NEW,
   WXTBAR_ADB_DELETE,
   WXTBAR_ADB_UNDO,
   WXTBAR_SEP,
   WXTBAR_ADB_FINDNEXT,
   -1
};

// this array stores all toolbar buttons for the given frame (the index in it
// is from the previous enum)
static const int *g_aFrameToolbars[WXFRAME_MAX] =
{
   g_aMainTbar,
   g_aComposeTbar,
   g_aFolderTbar,
   g_aMsgTbar,
   g_aAdbTbar
};


// array of descriptions of all menu items
// NB: by convention, if the menu item opens another window (or a dialog), it
//     should end with an ellipsis (`...')
static const MenuItemInfo g_aMenuItems[] =
{
   // filler for WXMENU_LAYOUT_CLICK
   { WXMENU_SEPARATOR,     "",                  ""                         },

   // file
   { WXMENU_FILE_OPEN,     gettext_noop("&Open Folder..."),   gettext_noop("Open an existing message folder")                  },
   { WXMENU_FILE_COMPOSE,  gettext_noop("&Compose Message"),  gettext_noop("Start a new message")      },
   { WXMENU_FILE_POST,  gettext_noop("Post &News Article"),   gettext_noop("Write a news article and post it")      },
   { WXMENU_SEPARATOR,     "",                  ""                         },
   { WXMENU_FILE_PRINT_SETUP,    gettext_noop("P&rint Setup"),     gettext_noop("Configure printing")  },
   { WXMENU_FILE_PAGE_SETUP,    gettext_noop("P&age Setup"),     gettext_noop("Configure page setup")  },
#ifdef USE_PS_PRINTING
   // extra postscript printing
   { WXMENU_FILE_PRINT_SETUP_PS,    gettext_noop("&Print PS Setup"),     gettext_noop("Configure PostScript printing")  },
//   { WXMENU_FILE_PAGE_SETUP_PS,    gettext_noop("P&age PS Setup"),     gettext_noop("Configure PostScript page setup")  },
#endif
   { WXMENU_SEPARATOR,     "",                  ""                         },
   { WXMENU_FILE_CREATE,   gettext_noop("Create &Folder..."), gettext_noop("Creates a new folder definition")               },

#ifdef USE_PYTHON
   { WXMENU_SEPARATOR,     "",                  ""                         },
   { WXMENU_FILE_SCRIPT,   gettext_noop("Run &Script..."),    gettext_noop("Run a simple python script")},
#endif // USE_PYTHON

   { WXMENU_SEPARATOR,     "",                  ""                         },
   { WXMENU_FILE_CLOSE,    gettext_noop("&Close Window"),     gettext_noop("Close this window")        },
   { WXMENU_SEPARATOR,     "",                  ""                         },
   { WXMENU_FILE_EXIT,     gettext_noop("E&xit"),             gettext_noop("Quit the application")     },

   // normal edit
   { WXMENU_EDIT_CUT,  gettext_noop("C&ut"), gettext_noop("Cut selection and copy it to clipboard")           },
   { WXMENU_EDIT_COPY, gettext_noop("&Copy"), gettext_noop("Copy selection to clipboard")           },
   { WXMENU_EDIT_PASTE,gettext_noop("&Paste"), gettext_noop("Paste from clipboard")           },
   { WXMENU_SEPARATOR,     "",                  ""                         },
   { WXMENU_EDIT_ADB,      gettext_noop("&Address books..."), gettext_noop("Edit the address book(s)") },
   { WXMENU_EDIT_PREF,     gettext_noop("Pre&ferences..."),   gettext_noop("Change options")           },
   { WXMENU_SEPARATOR,     "",                  ""                         },
   { WXMENU_EDIT_RESTORE_PREF,
                           gettext_noop("&Restore defaults..."), gettext_noop("Restore default options values") },
   { WXMENU_SEPARATOR,     "",                  ""                         },
   { WXMENU_EDIT_SAVE_PREF,gettext_noop("&Save Preferences"), gettext_noop("Save options")             },

   // msg
   { WXMENU_MSG_OPEN,      gettext_noop("&Open"),             gettext_noop("View selected message")    },
   { WXMENU_MSG_PRINT,     gettext_noop("&Print"),            gettext_noop("Print this message")       },
   { WXMENU_MSG_PRINT_PREVIEW, gettext_noop("Print Pre&view"),gettext_noop("Preview a printout of this message")       },
#ifdef USE_PS_PRINTING
   { WXMENU_MSG_PRINT_PS,     gettext_noop("PS-Prin&t"),      gettext_noop("Print this message as PostScript")       },
   { WXMENU_MSG_PRINT_PREVIEW_PS,     gettext_noop("PS&-Print Preview"),      gettext_noop("View PostScript printout")       },
#endif
   { WXMENU_MSG_REPLY,     gettext_noop("&Reply"),            gettext_noop("Reply to this message")    },
   { WXMENU_MSG_FORWARD,   gettext_noop("&Forward"),          gettext_noop("Forward this message")     },
   { WXMENU_SEPARATOR,     "",                  ""                         },
   { WXMENU_MSG_SAVE_TO_FILE, gettext_noop("&Save as File"),
     gettext_noop("Export message to a file")   },
   { WXMENU_MSG_SAVE_TO_FOLDER, gettext_noop("&Copy to Folder"),gettext_noop("Save message to another folder")   },
   { WXMENU_MSG_MOVE_TO_FOLDER, gettext_noop("&Move to Folder"),gettext_noop("Move message to another folder")   },
   { WXMENU_MSG_DELETE,    gettext_noop("&Delete"),           gettext_noop("Delete this message")      },
   { WXMENU_MSG_UNDELETE,  gettext_noop("&Undelete"),         gettext_noop("Undelete message")         },
   { WXMENU_MSG_EXPUNGE,   gettext_noop("Ex&punge"),          gettext_noop("Expunge")                  },
   { WXMENU_SEPARATOR,     "",                  ""                         },
   { WXMENU_MSG_SELECTALL, gettext_noop("Select &all"),       gettext_noop("Select all messages")      },
   { WXMENU_MSG_DESELECTALL,gettext_noop("D&eselect all"),    gettext_noop("Unselect all messages")    },
   { WXMENU_SEPARATOR,     "",                  ""                         },
   { WXMENU_MSG_TOGGLEHEADERS,gettext_noop("Show &header"),    gettext_noop("Toggle display of message header") },
   { WXMENU_MSG_SHOWRAWTEXT,  gettext_noop("Show ra&w message"), gettext_noop("Show the raw message text") },
   { WXMENU_MSG_FIND,  gettext_noop("F&ind..."), gettext_noop("Find text in message") },

   // compose
   { WXMENU_COMPOSE_INSERTFILE,
                           gettext_noop("&Insert file..."),   gettext_noop("Insert a file")            },
   { WXMENU_COMPOSE_SEND,  gettext_noop("&Send"),             gettext_noop("Send the message now")     },
   { WXMENU_COMPOSE_SEND_KEEP_OPEN,  gettext_noop("Send and &keep"),
     gettext_noop("Send the message now and keep the editor open")     },
   { WXMENU_COMPOSE_PRINT, gettext_noop("&Print"),            gettext_noop("Print the message")        },
   { WXMENU_SEPARATOR,     "",                  ""                         },
   { WXMENU_COMPOSE_SAVETEXT,gettext_noop("Save &text..."),   gettext_noop("Save message text to file")},
   { WXMENU_COMPOSE_LOADTEXT,gettext_noop("In&sert text..."), gettext_noop("Insert text file")         },
   { WXMENU_COMPOSE_CLEAR, gettext_noop("&Clear"),            gettext_noop("Delete message contents")  },
   { WXMENU_SEPARATOR,     "",                  ""                         },
   { WXMENU_COMPOSE_EXTEDIT, gettext_noop("&External editor"),gettext_noop("Invoke alternative editor")},

   // ADB book management
   { WXMENU_ADBBOOK_NEW,   gettext_noop("&New..."),           gettext_noop("Create a new address book")},
   { WXMENU_ADBBOOK_OPEN,  gettext_noop("&Open..."),          gettext_noop("Open an address book file")},
   { WXMENU_SEPARATOR,     "",                  ""                         },
#ifdef DEBUG
   { WXMENU_ADBBOOK_FLUSH, "&Flush",                  "Save changes to disk"                         },
   { WXMENU_SEPARATOR,     "",                  ""                         },
#endif // debug
   { WXMENU_ADBBOOK_PROP,  gettext_noop("&Properties..."), gettext_noop("View properties of the current address book")            },

   // ADB edit
   { WXMENU_ADBEDIT_NEW,   gettext_noop("&New entry..."),     gettext_noop("Create new entry/group")   },
   { WXMENU_ADBEDIT_DELETE,gettext_noop("&Delete"),           gettext_noop("Delete the selected items")},
   { WXMENU_ADBEDIT_RENAME,gettext_noop("&Rename..."),        gettext_noop("Rename the selected items")},
   { WXMENU_SEPARATOR,     "",                  ""                         },
   { WXMENU_ADBEDIT_CUT,   gettext_noop("Cu&t"),              gettext_noop("Copy and delete selected items")                    },
   { WXMENU_ADBEDIT_COPY,  gettext_noop("&Copy"),             gettext_noop("Copy selected items")      },
   { WXMENU_ADBEDIT_PASTE, gettext_noop("&Paste"),            gettext_noop("Paste here")               },
   { WXMENU_SEPARATOR,     "",                  ""                         },
   { WXMENU_ADBEDIT_UNDO,  gettext_noop("&Undo changes"),     gettext_noop("Undo all changes to the entry being edited")       },

   // ADB search
   { WXMENU_ADBFIND_FIND,  gettext_noop("&Find..."),          gettext_noop("Find entry by name or contents")                 },
   { WXMENU_ADBFIND_NEXT,  gettext_noop("Find &next"),        gettext_noop("Go to the next match")     },
   { WXMENU_SEPARATOR,     "",                  ""                         },
   { WXMENU_ADBFIND_GOTO,  gettext_noop("&Go To..."),         gettext_noop("Go to specified entry")    },

   // help
   { WXMENU_HELP_ABOUT,    gettext_noop("&About..."),         gettext_noop("Displays the program information and copyright")  },
   { WXMENU_HELP_RELEASE_NOTES,    gettext_noop("&Release Notes..."), gettext_noop("Displays notes about the current release.")  },
   { WXMENU_HELP_FAQ,    gettext_noop("&FAQ..."),         gettext_noop("Displays the list of Frequently Asked Questions.")  },
   { WXMENU_SEPARATOR,     "",                  ""                         },
   { WXMENU_HELP_CONTEXT, gettext_noop("&Help"),    gettext_noop("Help on current context...")},
   { WXMENU_HELP_CONTENTS, gettext_noop("Help &Contents"),    gettext_noop("Contents of help system...")},
   { WXMENU_HELP_SEARCH,   gettext_noop("&Search Help..."),      gettext_noop("Search help system for keyword...")},
   { WXMENU_HELP_COPYRIGHT,   gettext_noop("C&opyright"),      gettext_noop("Show Copyright.")}
};

// ============================================================================
// implementation
// ============================================================================

// ----------------------------------------------------------------------------
// menu stuff
// ----------------------------------------------------------------------------
void AppendToMenu(wxMenu *menu, int n)
{
   if ( g_aMenuItems[n].idMenu == WXMENU_SEPARATOR )
      menu->AppendSeparator();
   else {
      menu->Append(g_aMenuItems[n].idMenu,
                   wxGetTranslation(g_aMenuItems[n].label),
                   wxGetTranslation(g_aMenuItems[n].helpstring));
   }
}

void AppendToMenu(wxMenu *menu, int nFirst, int nLast)
{
   // consistency check which ensures (well, helps to ensure) that the array
   // and enum are in sync
   wxASSERT( WXSIZEOF(g_aMenuItems) == WXMENU_END + 1 );

   // in debug mode we also verify if the keyboard accelerators are ok
#ifdef DEBUG
   wxString strAccels;
#endif

   for ( int n = nFirst; n <= nLast; n++ ) {
#     ifdef DEBUG
      const char *label = wxGetTranslation(g_aMenuItems[n].label);
      if ( !IsEmpty(label) ) {
         const char *p = strchr(label, '&');
         if ( p == NULL ) {
            wxLogWarning("Menu label '%s' doesn't have keyboard accelerator.",
                         label);
         }
         else {
            char c = *++p;
            if ( strAccels.Find(c) != -1 ) {
               wxLogWarning("Duplicate accelerator %c (in '%s')", c, label);
            }

            strAccels += c;
         }
      }
      // else: it must be a separator

#     endif //DEBUG

      AppendToMenu(menu, n);
   }
}

// ----------------------------------------------------------------------------
// toolbar stuff
// ----------------------------------------------------------------------------

// add the given button to the toolbar
void AddToolbarButton(wxToolBar *toolbar, int nButton)
{
   if ( nButton == WXTBAR_SEP ) {
      toolbar->AddSeparator();
   }
   else {
      TB_AddTool(toolbar,
                 String(g_aToolBarData[nButton].bmp),
                 g_aToolBarData[nButton].id,
                 g_aToolBarData[nButton].tooltip);
   }
}

// add all buttons for the given frame to the toolbar
void AddToolbarButtons(wxToolBar *toolbar, wxFrameId frameId)
{
   wxASSERT( WXSIZEOF(g_aToolBarData) == WXTBAR_MAX );

#ifdef __WXMSW__
   // we use the icons of non standard size
   toolbar->SetToolBitmapSize(wxSize(32, 32));
#endif

   // if the buttons were of other size we'd have to use this function
   // (standard size is 16x15 under MSW)
   // toolbar->SetDefaultSize(wxSize(24, 24));

   wxASSERT( frameId < WXFRAME_MAX );
   const int *aTbarIcons = g_aFrameToolbars[frameId];

#ifdef __WXGTK__
   // it looks better like this under GTK
   AddToolbarButton(toolbar, WXTBAR_SEP);
#endif

   // first add the "Close" icon - but only if we're not the main frame
   if ( frameId != WXFRAME_MAIN ) {
      AddToolbarButton(toolbar, WXTBAR_CLOSE);
      AddToolbarButton(toolbar, WXTBAR_SEP);
   }

   for ( size_t nButton = 0; aTbarIcons[nButton] != -1 ; nButton++ ) {
      AddToolbarButton(toolbar, aTbarIcons[nButton]);
   }

   // next add the "Help" and "Exit" buttons
   AddToolbarButton(toolbar, WXTBAR_SEP);
   AddToolbarButton(toolbar, WXTBAR_MAIN_HELP);
   AddToolbarButton(toolbar, WXTBAR_SEP);
   AddToolbarButton(toolbar, WXTBAR_MAIN_EXIT);

   // must do it for the toolbar to be shown properly
#ifdef __WXMSW__
   toolbar->CreateTools();
#elif defined(__WXGTK__)
   toolbar->Realize();
   toolbar->Layout();
#endif // Windows
}

/* vi: set tw=0 */
